from tiviwiki import app
from flask import request, flash, redirect, url_for, escape, session, abort, render_template
from functools import wraps
from utils import hashPassword, WikiToHTML
from database import query_db, affected_rows

import forms

# -----------------------------------------------------------------------------
# Decorators: force user to be either logged in or logged out
# -----------------------------------------------------------------------------

def login_required(func):
    @wraps(func)
    def decorated_function(*args, **kwargs):
        if not 'logged_in' in session:
            return redirect(url_for('login', next=request.url))
        return func(*args, **kwargs)
    return decorated_function

def logout_required(func):
    @wraps(func)
    def decorated_function(*args, **kwargs):
        if 'logged_in' in session:
            flash('You cannot open this page while you are logged in!', 'error')
            return redirect(url_for('index'))
        return func(*args, **kwargs)
    return decorated_function

def admin_required(func):
    @wraps(func)
    def decorated_function(*args, **kwargs):
        if not 'logged_in' in session:
            return redirect(url_for('login', next=request.url))
        if session['privileges'] != 'admin':
            abort(403)
        return func(*args, **kwargs)
    return decorated_function

# -----------------------------------------------------------------------------
# Entry Point
# -----------------------------------------------------------------------------

@app.route('/')
def index():
    return render_template('index.html')

# -----------------------------------------------------------------------------
# Login / Logout / Register pages
# -----------------------------------------------------------------------------

@app.route('/login', methods=['GET', 'POST'])
@logout_required
def login():
    error = None

    form = forms.LoginForm(request.form)
    if request.method == 'POST' and form.validate():
        rows = query_db("""SELECT `id`, `username`, `privileges`
                           FROM users
                           WHERE `username`=%s and `password`=%s""",
                        (form.username.data, hashPassword(form.password.data),))

        if rows!= []:
            row = rows[0]
            username = row['username']
            flash('Welcome back <strong>%s</strong>, you were successfully logged in' % escape(username),
                  'success')
            session['username'] = username
            session['privileges'] = row['privileges']
            session['user_id'] = row['id']
            session['logged_in'] = True

            return redirect(request.args.get('next') or url_for('index'))
        else:
            error = "Invalid credentials"

    return render_template('login.html', error=error, form=form)

@app.route('/logout')
@login_required
def logout():
    session.clear()
    flash('You were logged out', 'info')
    return redirect(url_for('index'))

@app.route('/register', methods=['GET', 'POST'])
@logout_required
def register():
    errors = []

    form = forms.RegistrationForm(request.form)
    if request.method == 'POST' and form.validate():
        rows = query_db("SELECT * FROM users WHERE `username`=%s or `email`=%s",
                        (form.username.data, form.email.data,))
        if rows == []:
            query_db("""INSERT INTO users (`username`, `password`, `email`)
                        VALUES (%s, %s, %s)""", (form.username.data,
                                                 hashPassword(form.password.data),
                                                 form.email.data,))

            flash('Thank you for registering, <strong>%s</strong>!' % escape(form.username.data),
                  'success')
            return redirect(url_for('login'))
        else:
            for row in rows:
                if row['username'] == form.username.data:
                    errors += ['Username <strong>%s</strong> is already used' % escape(form.username.data)]
                if row['email'] == form.email.data:
                    errors += ['Email <strong>%s</strong> is already used' % escape(form.email.data)]

    return render_template('register.html', errors=errors, form=form)

# -----------------------------------------------------------------------------
# Wiki interaction: add/edit/delete/publish/list articles
# -----------------------------------------------------------------------------

@app.route('/wiki/add', methods=['GET', 'POST'])
@login_required
def wiki_add_article():
    errors = []
    preview = None

    form = forms.WikiAddPage(request.form)
    if request.method == 'POST' and form.validate():
        if request.form['submit'] == 'add':
            if query_db("SELECT 1 FROM articles WHERE `title`=%s", (form.title.data,)) == []:
                query_db("""INSERT INTO articles (`title`, `content`, `user_id`)
                            VALUES (%s, %s, %s)""", (form.title.data,
                                                     form.content.data,
                                                     session['user_id'],))

                flash('Your article was added successfully!', 'success')
                return redirect(url_for('wiki_user_articles'))
            else:
                errors += ['Another article with the same title exists']
        else:
            preview = WikiToHTML(form.content.data)

    return render_template('wiki/add_article.html', errors=errors, form=form, preview=preview)

@app.route('/wiki/edit/<int:article_id>', methods=['GET', 'POST'])
@login_required
def wiki_edit_article(article_id):
    errors = []
    preview = None

    form = forms.WikiAddPage(request.form)
    if request.method == 'POST' and form.validate():
        if request.form['submit'] == 'edit':
            try:
                query = """UPDATE articles SET `title`=%s, `content`=%s, `lastupdate_date`=now(), `status`='draft'
                           WHERE `id`=%s AND `user_id`=%s AND `status`!='published'"""
                query_db(query, (form.title.data, form.content.data, article_id, session['user_id'],))
                if affected_rows() == 0:
                    flash('You are not allowed to edit this article', 'error')
                    return redirect(url_for('index'))

                flash('Your article was edited successfully!', 'success')
                return redirect(url_for('wiki_user_articles'))
            except:
                flash('Another article with the same name already exists!', 'error')
        else:
            preview = WikiToHTML(form.content.data)
    elif request.method == 'GET':
        rows = query_db("""SELECT`title`, `content`
                           FROM articles
                           WHERE `id`=%s AND `user_id`=%s AND `status` != 'published'""",
                        (article_id, session['user_id'],))

        if rows == []:
            flash("This article does not exist or you are not allowed to read it", "error")
            abort(404)

        article = rows[0]
        form.title.data = article['title']
        form.content.data = article['content']

    return render_template('wiki/edit_article.html', errors=errors, form=form, preview=preview)

@app.route('/wiki/delete/<int:article_id>')
@login_required
def wiki_delete_article(article_id):
    query = """DELETE FROM articles WHERE `id`=%s AND `user_id`=%s AND `status`!='published'"""
    query_db(query, (article_id, session['user_id'],))

    if affected_rows() == 0:
        flash('You are not allowed to delete this article.', 'error')
        return redirect(url_for('wiki_user_articles'))

    flash('Your article was successfully deleted.', 'success')
    return redirect(url_for('wiki_user_articles'))

@app.route('/wiki/publish/<int:article_id>')
@login_required
def wiki_publish_article(article_id):
    query = """UPDATE articles SET `status`='waiting'
               WHERE `id`=%s AND `user_id`=%s AND `status`!='published'"""
    query_db(query, (article_id, session['user_id'],))

    if affected_rows() == 0:
        flash('You are not allowed to publish this article, or it is already published.', 'error')
        return redirect(url_for('wiki_user_articles'))

    flash("Thank you, your article is now waiting for validation by an administrator. You'll receive a notice when this is done.", "success")
    return redirect(url_for('wiki_user_articles'))

@app.route('/wiki/articles')
@login_required
def wiki_articles():
    articles = query_db("""SELECT articles.`title`, articles.`id`, users.`username`,
                                  IF(articles.`lastupdate_date` is NULL,
                                     articles.`creation_date`,
                                     articles.`lastupdate_date`) as date
                           FROM articles
                           INNER JOIN users
                           ON articles.`user_id`=users.`id`
                           WHERE articles.`status`='published'
                           ORDER BY `creation_date` DESC""")

    return render_template('wiki/articles.html', articles=articles)

@app.route('/wiki/article/<int:article_id>')
@login_required
def wiki_article(article_id):
    rows = query_db("""SELECT articles.`title`, articles.`content`, articles.`creation_date`, users.`username`
                       FROM articles
                       INNER JOIN users
                       ON articles.`user_id`=users.`id`
                       WHERE articles.`id`=%s AND (articles.`status`='published' OR
                                                   articles.`user_id`=%s OR
                                                   %s='admin')""",
                    (article_id, session['user_id'], session['privileges'],))

    if rows == []:
        flash("This article does not exist or you are not allowed to read it", "error")
        abort(404)

    article = rows[0]
    article['content'] = WikiToHTML(article['content'])

    return render_template('wiki/article.html', article=article)

# -----------------------------------------------------------------------------
# User articles / settings / notifications
# -----------------------------------------------------------------------------

@app.route('/wiki/user/articles')
@login_required
def wiki_user_articles():

    user_articles = query_db("""SELECT articles.`title`, articles.`id`, articles.`status`,
                                       articles.`creation_date`, articles.`lastupdate_date`
                                FROM articles
                                INNER JOIN users
                                ON articles.`user_id`=users.`id`
                                WHERE articles.`user_id`=%s
                                ORDER BY `creation_date` DESC""", (session['user_id'],))

    return render_template('wiki/user_articles.html', articles=user_articles)

@app.route('/wiki/user/settings', methods=['GET', 'POST'])
@login_required
def wiki_user_settings():
    form = forms.UserSettings(request.form)
    if request.method == 'POST' and form.validate():
        email = form.email.data
        password = form.password.data

        if password != u'':
            query_db("""UPDATE users SET `email`=%s, `password`=%s WHERE `id`=%s""", (email,
                                                                                      hashPassword(password),
                                                                                      session['user_id']))
        else:
            query_db("""UPDATE users SET `email`=%s WHERE `id`=%s""", (email, session['user_id'],))

        flash('Your settings were successfully edited', 'success')
    else:
        if form.email.data is None:
            rows = query_db("SELECT `email` FROM `users` WHERE `id`=%s", (session['user_id'],))
            form.email.data = rows[0]['email']

    return render_template('wiki/user_settings.html', form=form)

@app.route('/wiki/user/notices')
@login_required
def wiki_user_notices():
    notices = query_db("""SELECT notices.`id`, notices.`content`, notices.`date`,
                                 users.`username` AS from_user
                          FROM notices
                          INNER JOIN users
                          ON users.`id`=notices.`from_user_id`
                          WHERE notices.`to_user_id`=%s""", (session['user_id'],))

    return render_template('wiki/user_notices.html', notices=notices)

@app.route('/wiki/user/notices/delete/<int:notice_id>')
@login_required
def wiki_user_delete_notice(notice_id):
    query = """DELETE FROM notices WHERE `id`=%s AND `to_user_id`=%s"""
    query_db(query, (notice_id, session['user_id'],))

    if affected_rows() == 0:
        flash('You are not allowed to to this.', 'error')
        return redirect(url_for('wiki_user_notices'))

    flash('Your notice wass successfully deleted.', 'success')
    return redirect(url_for('wiki_user_notices'))

# -----------------------------------------------------------------------------
# Administration
# -----------------------------------------------------------------------------

@app.route('/wiki/admin/templates')
@admin_required
def admin_templates():
    # You don't need this source code at the moment ;)
    return render_template('admin/templates.html')

@app.route('/wiki/admin/articles')
@admin_required
def admin_articles():
    # You don't need this source code at the moment ;)
    return render_template('admin/articles.html')

# -----------------------------------------------------------------------------
# Error Handlers
# -----------------------------------------------------------------------------

@app.errorhandler(404)
def page_not_found(error):
    return render_template('page_not_found.html'), 404
